/* ReaPack: Package manager for REAPER
 * Copyright (C) 2015-2025  Christian Fillion
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef REAPACK_SERIALIZER_HPP
#define REAPACK_SERIALIZER_HPP

#include <array>
#include <string>
#include <list>

class Serializer {
public:
  typedef std::array<int, 2> Record;
  typedef std::list<Record> Data;

  Serializer() : m_userVersion(0) {}
  int userVersion() const { return m_userVersion; }
  operator bool() const { return m_userVersion > 0; }

  Data read(const std::string &, int userVersion);
  std::string write(const Data &in) const;

private:
  int m_userVersion;
};

#endif
